@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -69,6 +69,7 @@ static DBusConnection *connection;
 
 static GHashTable *bluetooth_devices = NULL;
 static GHashTable *bluetooth_networks = NULL;
+static GHashTable *pending_networks = NULL;
 
 static int pan_probe(struct connman_network *network)
 {
@@ -271,7 +272,7 @@ static struct connman_network_driver pan_driver = {
 	.disconnect	= pan_disconnect,
 };
 
-static gboolean network_changed(DBusConnection *connection,
+static gboolean network_changed(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -446,7 +447,7 @@ static void network_properties_reply(DBusPendingCall *call, void *user_data)
 
 	connman_network_set_group(network, ident);
 
-	g_hash_table_insert(bluetooth_networks, g_strdup(path), network);
+	g_hash_table_replace(bluetooth_networks, g_strdup(path), network);
 
 done:
 	dbus_message_unref(reply);
@@ -454,7 +455,7 @@ done:
 	dbus_pending_call_unref(call);
 }
 
-static void add_network(struct connman_device *device, const char *path)
+static void add_network(const char *path)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
@@ -486,8 +487,7 @@ done:
 	dbus_message_unref(message);
 }
 
-static void check_networks(struct connman_device *device,
-						DBusMessageIter *array)
+static void check_networks(DBusMessageIter *array)
 {
 	DBusMessageIter value;
 
@@ -501,13 +501,30 @@ static void check_networks(struct connman_device *device,
 
 		dbus_message_iter_get_basic(&value, &path);
 
-		add_network(device, path);
+		add_network(path);
 
 		dbus_message_iter_next(&value);
 	}
 }
 
-static gboolean adapter_changed(DBusConnection *connection,
+static void check_pending_networks(const char *adapter)
+{
+	GSList *networks, *list;
+
+	networks = g_hash_table_lookup(pending_networks, adapter);
+	if (networks == NULL)
+		return;
+
+	for (list = networks; list != NULL; list = list->next) {
+		char *path = list->data;
+
+		add_network(path);
+	}
+
+	g_hash_table_remove(pending_networks, adapter);
+}
+
+static gboolean adapter_changed(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -534,22 +551,24 @@ static gboolean adapter_changed(DBusConnection *connection,
 
 		dbus_message_iter_get_basic(&value, &val);
 		connman_device_set_powered(device, val);
+		if (val == TRUE)
+			check_pending_networks(path);
 	} else if (g_str_equal(key, "Discovering") == TRUE) {
 		dbus_bool_t val;
 
 		dbus_message_iter_get_basic(&value, &val);
 		connman_device_set_scanning(device, val);
 	} else if (g_str_equal(key, "Devices") == TRUE) {
-		check_networks(device, &value);
+		check_networks(&value);
 	}
 
 	return TRUE;
 }
 
-static gboolean device_removed(DBusConnection *connection,
+static gboolean device_removed(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
-	const char *network_path, *identifier;
+	const char *network_path;
 	struct connman_network *network;
 	struct connman_device *device;
 	DBusMessageIter iter;
@@ -569,16 +588,12 @@ static gboolean device_removed(DBusConnection *connection,
 	if (device == NULL)
 		return TRUE;
 
-	identifier = connman_network_get_identifier(network);
-
 	g_hash_table_remove(bluetooth_networks, network_path);
 
-	connman_device_remove_network(device, identifier);
-
 	return TRUE;
 }
 
-static gboolean device_changed(DBusConnection *connection,
+static gboolean device_changed(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -598,11 +613,67 @@ static gboolean device_changed(DBusConnection *connection,
 	DBG("key %s", key);
 
 	if (g_str_equal(key, "UUIDs") == TRUE)
-		add_network(NULL, path);
+		add_network(path);
 
 	return TRUE;
 }
 
+static void remove_device_networks(struct connman_device *device)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+	GSList *key_list = NULL;
+	GSList *list;
+
+	if (bluetooth_networks == NULL)
+		return;
+
+	g_hash_table_iter_init(&iter, bluetooth_networks);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		struct connman_network *network = value;
+
+		if (connman_network_get_device(network) != device)
+			continue;
+
+		key_list = g_slist_append(key_list, key);
+	}
+
+	for (list = key_list; list != NULL; list = list->next) {
+		const char *network_path = list->data;
+
+		g_hash_table_remove(bluetooth_networks, network_path);
+	}
+
+	g_slist_free(key_list);
+}
+
+static void add_pending_networks(const char *adapter, DBusMessageIter *array)
+{
+	DBusMessageIter value;
+	GSList *list = NULL;
+
+	if (dbus_message_iter_get_arg_type(array) != DBUS_TYPE_ARRAY)
+		return;
+
+	dbus_message_iter_recurse(array, &value);
+
+	while (dbus_message_iter_get_arg_type(&value) == DBUS_TYPE_OBJECT_PATH) {
+		const char *path;
+
+		dbus_message_iter_get_basic(&value, &path);
+
+		list = g_slist_prepend(list, g_strdup(path));
+
+		dbus_message_iter_next(&value);
+	}
+
+	if (list == NULL)
+		return;
+
+	g_hash_table_replace(pending_networks, g_strdup(adapter), list);
+}
+
 static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 {
 	char *path = user_data;
@@ -634,6 +705,8 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 	if (device != NULL)
 		goto update;
 
+	g_hash_table_insert(bluetooth_devices, g_strdup(path), device);
+
 	ether_aton_r(address, &addr);
 
 	snprintf(ident, 13, "%02x%02x%02x%02x%02x%02x",
@@ -654,11 +727,10 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 
 	if (connman_device_register(device) < 0) {
 		connman_device_unref(device);
+		g_hash_table_remove(bluetooth_devices, path);
 		goto done;
 	}
 
-	g_hash_table_insert(bluetooth_devices, g_strdup(path), device);
-
 update:
 	connman_device_set_string(device, "Address", address);
 	connman_device_set_string(device, "Name", name);
@@ -667,8 +739,11 @@ update:
 	connman_device_set_powered(device, powered);
 	connman_device_set_scanning(device, scanning);
 
-	if (powered == TRUE)
-		check_networks(device, &networks);
+	if (powered == FALSE) {
+		remove_device_networks(device);
+		add_pending_networks(path, &networks);
+	} else
+		check_networks(&networks);
 
 done:
 	dbus_message_unref(reply);
@@ -676,7 +751,7 @@ done:
 	dbus_pending_call_unref(call);
 }
 
-static void add_adapter(DBusConnection *connection, const char *path)
+static void add_adapter(DBusConnection *conn, const char *path)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
@@ -690,7 +765,7 @@ static void add_adapter(DBusConnection *connection, const char *path)
 
 	dbus_message_set_auto_start(message, FALSE);
 
-	if (dbus_connection_send_with_reply(connection, message,
+	if (dbus_connection_send_with_reply(conn, message,
 						&call, TIMEOUT) == FALSE) {
 		connman_error("Failed to get adapter properties for %s", path);
 		goto done;
@@ -708,32 +783,33 @@ done:
 	dbus_message_unref(message);
 }
 
-static gboolean adapter_added(DBusConnection *connection, DBusMessage *message,
+static gboolean adapter_added(DBusConnection *conn, DBusMessage *message,
 				void *user_data)
 {
 	const char *path;
 
 	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
 				DBUS_TYPE_INVALID);
-	add_adapter(connection, path);
+	add_adapter(conn, path);
 	return TRUE;
 }
 
-static void remove_adapter(DBusConnection *connection, const char *path)
+static void remove_adapter(DBusConnection *conn, const char *path)
 {
 	DBG("path %s", path);
 
 	g_hash_table_remove(bluetooth_devices, path);
+	g_hash_table_remove(pending_networks, path);
 }
 
-static gboolean adapter_removed(DBusConnection *connection, DBusMessage *message,
+static gboolean adapter_removed(DBusConnection *conn, DBusMessage *message,
 				void *user_data)
 {
 	const char *path;
 
 	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
 				DBUS_TYPE_INVALID);
-	remove_adapter(connection, path);
+	remove_adapter(conn, path);
 	return TRUE;
 }
 
@@ -785,22 +861,48 @@ static void unregister_device(gpointer data)
 
 	DBG("");
 
+	remove_device_networks(device);
+
 	connman_device_unregister(device);
 	connman_device_unref(device);
 }
 
-static void bluetooth_connect(DBusConnection *connection, void *user_data)
+static void remove_network(gpointer data)
+{
+	struct connman_network *network = data;
+	struct connman_device *device;
+
+	DBG("network %p", network);
+
+	device = connman_network_get_device(network);
+	if (device != NULL)
+		connman_device_remove_network(device, network);
+
+	connman_network_unref(network);
+}
+
+static void remove_pending_networks(gpointer data)
+{
+	GSList *list = data;
+
+	g_slist_free_full(list, g_free);
+}
+
+static void bluetooth_connect(DBusConnection *conn, void *user_data)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
 
-	DBG("connection %p", connection);
+	DBG("connection %p", conn);
 
 	bluetooth_devices = g_hash_table_new_full(g_str_hash, g_str_equal,
 						g_free, unregister_device);
 
 	bluetooth_networks = g_hash_table_new_full(g_str_hash, g_str_equal,
-						g_free, NULL);
+						g_free, remove_network);
+
+	pending_networks = g_hash_table_new_full(g_str_hash, g_str_equal,
+					g_free, remove_pending_networks);
 
 	message = dbus_message_new_method_call(BLUEZ_SERVICE, "/",
 				BLUEZ_MANAGER_INTERFACE, LIST_ADAPTERS);
@@ -809,7 +911,7 @@ static void bluetooth_connect(DBusConnection *connection, void *user_data)
 
 	dbus_message_set_auto_start(message, FALSE);
 
-	if (dbus_connection_send_with_reply(connection, message,
+	if (dbus_connection_send_with_reply(conn, message,
 						&call, TIMEOUT) == FALSE) {
 		connman_error("Failed to get Bluetooth adapters");
 		goto done;
@@ -826,23 +928,41 @@ done:
 	dbus_message_unref(message);
 }
 
-static void bluetooth_disconnect(DBusConnection *connection, void *user_data)
+static void bluetooth_disconnect(DBusConnection *conn, void *user_data)
 {
-	DBG("connection %p", connection);
+	DBG("connection %p", conn);
 
 	if (bluetooth_devices == NULL)
 		return;
 
 	g_hash_table_destroy(bluetooth_networks);
+	bluetooth_networks = NULL;
 	g_hash_table_destroy(bluetooth_devices);
 	bluetooth_devices = NULL;
+	g_hash_table_destroy(pending_networks);
+	pending_networks = NULL;
 }
 
 static int bluetooth_probe(struct connman_device *device)
 {
+	GHashTableIter iter;
+	gpointer key, value;
+
 	DBG("device %p", device);
 
-	return 0;
+	if (bluetooth_devices == NULL)
+		return -ENOTSUP;
+
+	g_hash_table_iter_init(&iter, bluetooth_devices);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		struct connman_device *device_pan = value;
+
+		if (device == device_pan)
+			return 0;
+	}
+
+	return -ENOTSUP;
 }
 
 static void bluetooth_remove(struct connman_device *device)
@@ -875,7 +995,7 @@ static void powered_reply(DBusPendingCall *call, void *user_data)
 	add_adapter(connection, user_data);
 }
 
-static int change_powered(DBusConnection *connection, const char *path,
+static int change_powered(DBusConnection *conn, const char *path,
 							dbus_bool_t powered)
 {
 	DBusMessage *message;
@@ -898,7 +1018,7 @@ static int change_powered(DBusConnection *connection, const char *path,
 	connman_dbus_property_append_basic(&iter, "Powered",
 						DBUS_TYPE_BOOLEAN, &powered);
 
-	if (dbus_connection_send_with_reply(connection, message,
+	if (dbus_connection_send_with_reply(conn, message,
 						&call, TIMEOUT) == FALSE) {
 		connman_error("Failed to change Powered property");
 		dbus_message_unref(message);
@@ -1134,33 +1254,34 @@ static int bluetooth_init(void)
 	watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
 			bluetooth_connect, bluetooth_disconnect, NULL, NULL);
 
-	added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	added_watch = g_dbus_add_signal_watch(connection, BLUEZ_SERVICE, NULL,
 						BLUEZ_MANAGER_INTERFACE,
 						ADAPTER_ADDED, adapter_added,
 						NULL, NULL);
 
-	removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	removed_watch = g_dbus_add_signal_watch(connection, BLUEZ_SERVICE, NULL,
 						BLUEZ_MANAGER_INTERFACE,
 						ADAPTER_REMOVED, adapter_removed,
 						NULL, NULL);
 
-	adapter_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
-						BLUEZ_ADAPTER_INTERFACE,
+	adapter_watch = g_dbus_add_signal_watch(connection, BLUEZ_SERVICE,
+						NULL, BLUEZ_ADAPTER_INTERFACE,
 						PROPERTY_CHANGED, adapter_changed,
 						NULL, NULL);
 
-	device_removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	device_removed_watch = g_dbus_add_signal_watch(connection,
+						BLUEZ_SERVICE, NULL,
 						BLUEZ_ADAPTER_INTERFACE,
 						DEVICE_REMOVED, device_removed,
 						NULL, NULL);
 
-	device_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	device_watch = g_dbus_add_signal_watch(connection, BLUEZ_SERVICE, NULL,
 						BLUEZ_DEVICE_INTERFACE,
 						PROPERTY_CHANGED, device_changed,
 						NULL, NULL);
 
-	network_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
-						BLUEZ_NETWORK_INTERFACE,
+	network_watch = g_dbus_add_signal_watch(connection, BLUEZ_SERVICE,
+						NULL, BLUEZ_NETWORK_INTERFACE,
 						PROPERTY_CHANGED, network_changed,
 						NULL, NULL);
 
@@ -1186,7 +1307,7 @@ static int bluetooth_init(void)
 	if (err < 0) {
 		connman_device_driver_unregister(&bluetooth_driver);
 		connman_network_driver_unregister(&pan_driver);
-		return err;
+		goto remove;
 	}
 
 	return 0;
