@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *  Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
  *  Copyright (C) 2011  BWM Car IT GmbH. All rights reserved.
  *
@@ -43,8 +43,6 @@
 
 #include "mcc.h"
 
-#define uninitialized_var(x) x = x
-
 #define OFONO_SERVICE			"org.ofono"
 
 #define OFONO_MANAGER_INTERFACE		OFONO_SERVICE ".Manager"
@@ -172,6 +170,7 @@ struct modem_data {
 	/* ConnectionContext Interface */
 	connman_bool_t active;
 	connman_bool_t set_active;
+	connman_bool_t valid_apn; /* APN is 'valid' if length > 0 */
 
 	/* SimManager Interface */
 	char *imsi;
@@ -257,51 +256,65 @@ static void network_context_free(struct network_context *context)
 
 static void set_connected(struct modem_data *modem)
 {
+	struct connman_service *service;
 	connman_bool_t setip = FALSE;
+	enum connman_ipconfig_method method;
+	char *nameservers;
+	int index;
 
 	DBG("%s", modem->path);
 
-	connman_network_set_index(modem->network, modem->context->index);
+	index = modem->context->index;
 
-	switch (modem->context->ipv4_method) {
-	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
-	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_MANUAL:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
-		break;
-
-	case CONNMAN_IPCONFIG_METHOD_FIXED:
-		connman_network_set_ipv4_method(modem->network,
-						modem->context->ipv4_method);
-		connman_network_set_ipaddress(modem->network,
-						modem->context->ipv4_address);
-		connman_network_set_nameservers(modem->network,
-					modem->context->ipv4_nameservers);
-		setip = TRUE;
-		break;
-
-	case CONNMAN_IPCONFIG_METHOD_DHCP:
-		connman_network_set_ipv4_method(modem->network,
-						modem->context->ipv4_method);
-		setip = TRUE;
-		break;
+	if (index < 0 || modem->context->ipv4_address == NULL) {
+		connman_error("Invalid index and/or address");
+		return;
 	}
 
-	switch (modem->context->ipv6_method) {
-	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
-	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_MANUAL:
-	case CONNMAN_IPCONFIG_METHOD_DHCP:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
-		break;
+	service = connman_service_lookup_from_network(modem->network);
+	if (service == NULL)
+		return;
+
+	method = modem->context->ipv4_method;
+	if (method == CONNMAN_IPCONFIG_METHOD_FIXED ||
+			method == CONNMAN_IPCONFIG_METHOD_DHCP)
+	{
+		connman_service_create_ip4config(service, index);
+		connman_network_set_index(modem->network, index);
+
+		connman_network_set_ipv4_method(modem->network, method);
 
-	case CONNMAN_IPCONFIG_METHOD_FIXED:
-		connman_network_set_ipv6_method(modem->network,
-							modem->context->ipv6_method);
-		connman_network_set_ipaddress(modem->network,
-							modem->context->ipv6_address);
 		setip = TRUE;
-		break;
+	}
+
+	if (method == CONNMAN_IPCONFIG_METHOD_FIXED) {
+		connman_network_set_ipaddress(modem->network,
+						modem->context->ipv4_address);
+	}
+
+	method = modem->context->ipv6_method;
+	if (method == CONNMAN_IPCONFIG_METHOD_FIXED) {
+		connman_service_create_ip6config(service, index);
+		connman_network_set_ipv6_method(modem->network, method);
+		connman_network_set_ipaddress(modem->network,
+						modem->context->ipv6_address);
+		setip = TRUE;
+	}
+
+	/* Set the nameservers */
+	if (modem->context->ipv4_nameservers != NULL &&
+			modem->context->ipv6_nameservers != NULL) {
+		nameservers = g_strdup_printf("%s %s",
+					modem->context->ipv4_nameservers,
+					modem->context->ipv6_nameservers);
+		connman_network_set_nameservers(modem->network, nameservers);
+		g_free(nameservers);
+	} else if (modem->context->ipv4_nameservers != NULL) {
+		connman_network_set_nameservers(modem->network,
+					modem->context->ipv4_nameservers);
+	} else if (modem->context->ipv6_nameservers != NULL) {
+		connman_network_set_nameservers(modem->network,
+					modem->context->ipv6_nameservers);
 	}
 
 	if (setip == TRUE)
@@ -312,6 +325,9 @@ static void set_disconnected(struct modem_data *modem)
 {
 	DBG("%s", modem->path);
 
+	if (modem->network == NULL)
+		return;
+
 	connman_network_set_connected(modem->network, FALSE);
 }
 
@@ -830,7 +846,7 @@ static void extract_ipv6_settings(DBusMessageIter *array,
 {
 	DBusMessageIter dict;
 	char *address = NULL, *gateway = NULL;
-	unsigned char prefix_length;
+	unsigned char prefix_length = 0;
 	char *nameservers = NULL;
 	const char *interface = NULL;
 	int index = -1;
@@ -933,7 +949,7 @@ static connman_bool_t ready_to_create_device(struct modem_data *modem)
 static void create_device(struct modem_data *modem)
 {
 	struct connman_device *device;
-	char *uninitialized_var(ident);
+	char *ident = NULL;
 
 	DBG("%s", modem->path);
 
@@ -1011,8 +1027,6 @@ static void add_network(struct modem_data *modem)
 	connman_network_set_string(modem->network, "Path",
 					modem->context->path);
 
-	connman_network_set_index(modem->network, modem->context->index);
-
 	if (modem->name != NULL)
 		connman_network_set_name(modem->network, modem->name);
 	else
@@ -1033,12 +1047,6 @@ static void add_network(struct modem_data *modem)
 		modem->network = NULL;
 		return;
 	}
-
-	/*
-	 * Create the ipconfig layer before trying to connect. Withouth
-	 * the ipconfig layer the core is not ready to process errors.
-	 */
-	connman_network_set_index(modem->network, -1);
 }
 
 static void remove_network(struct modem_data *modem)
@@ -1058,7 +1066,7 @@ static void remove_network(struct modem_data *modem)
 static int add_cm_context(struct modem_data *modem, const char *context_path,
 				DBusMessageIter *dict)
 {
-	const char *context_type;
+	const char *context_type = NULL;
 	struct network_context *context = NULL;
 	connman_bool_t active = FALSE;
 
@@ -1103,8 +1111,17 @@ static int add_cm_context(struct modem_data *modem, const char *context_path,
 			dbus_message_iter_get_basic(&value, &active);
 
 			DBG("%s Active %d", modem->path, active);
-		}
+		} else if (g_str_equal(key, "AccessPointName") == TRUE) {
+			const char *apn;
 
+			dbus_message_iter_get_basic(&value, &apn);
+			if (apn != NULL && strlen(apn) > 0)
+				modem->valid_apn = TRUE;
+			else
+				modem->valid_apn = FALSE;
+
+			DBG("%s AccessPointName '%s'", modem->path, apn);
+		}
 		dbus_message_iter_next(dict);
 	}
 
@@ -1118,6 +1135,12 @@ static int add_cm_context(struct modem_data *modem, const char *context_path,
 
 	g_hash_table_replace(context_hash, g_strdup(context_path), modem);
 
+	if (modem->valid_apn == TRUE && modem->attached == TRUE &&
+			has_interface(modem->interfaces,
+				OFONO_API_NETREG) == TRUE) {
+		add_network(modem);
+	}
+
 	return 0;
 }
 
@@ -1134,9 +1157,14 @@ static void remove_cm_context(struct modem_data *modem,
 
 	network_context_free(modem->context);
 	modem->context = NULL;
+
+	modem->valid_apn = FALSE;
+
+	if (modem->network != NULL)
+		remove_network(modem);
 }
 
-static gboolean context_changed(DBusConnection *connection,
+static gboolean context_changed(DBusConnection *conn,
 				DBusMessage *message,
 				void *user_data)
 {
@@ -1181,6 +1209,39 @@ static gboolean context_changed(DBusConnection *connection,
 			set_connected(modem);
 		else
 			set_disconnected(modem);
+	} else if (g_str_equal(key, "AccessPointName") == TRUE) {
+		const char *apn;
+
+		dbus_message_iter_get_basic(&value, &apn);
+
+		DBG("%s AccessPointName %s", modem->path, apn);
+
+		if (apn != NULL && strlen(apn) > 0) {
+			modem->valid_apn = TRUE;
+
+			if (modem->network != NULL)
+				return TRUE;
+
+			if (modem->attached == FALSE)
+				return TRUE;
+
+			if (has_interface(modem->interfaces,
+					OFONO_API_NETREG) == FALSE) {
+				return TRUE;
+			}
+
+			add_network(modem);
+
+			if (modem->active == TRUE)
+				set_connected(modem);
+		} else {
+			modem->valid_apn = FALSE;
+
+			if (modem->network == NULL)
+				return TRUE;
+
+			remove_network(modem);
+		}
 	}
 
 	return TRUE;
@@ -1272,7 +1333,7 @@ static int cm_get_contexts(struct modem_data *modem)
 	return -EINPROGRESS;
 }
 
-static gboolean cm_context_added(DBusConnection *connection,
+static gboolean cm_context_added(DBusConnection *conn,
 					DBusMessage *message,
 					void *user_data)
 {
@@ -1301,7 +1362,7 @@ static gboolean cm_context_added(DBusConnection *connection,
 	return TRUE;
 }
 
-static gboolean cm_context_removed(DBusConnection *connection,
+static gboolean cm_context_removed(DBusConnection *conn,
 					DBusMessage *message,
 					void *user_data)
 {
@@ -1445,7 +1506,7 @@ static void netreg_update_regdom(struct modem_data *modem,
 		connman_technology_set_regdom(alpha2);
 }
 
-static gboolean netreg_changed(DBusConnection *connection, DBusMessage *message,
+static gboolean netreg_changed(DBusConnection *conn, DBusMessage *message,
 				void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -1519,7 +1580,8 @@ static void netreg_properties_reply(struct modem_data *modem,
 		return;
 	}
 
-	add_network(modem);
+	if (modem->valid_apn == TRUE)
+		add_network(modem);
 
 	if (modem->active == TRUE)
 		set_connected(modem);
@@ -1553,7 +1615,7 @@ static void add_cdma_network(struct modem_data *modem)
 		set_connected(modem);
 }
 
-static gboolean cdma_netreg_changed(DBusConnection *connection,
+static gboolean cdma_netreg_changed(DBusConnection *conn,
 					DBusMessage *message,
 					void *user_data)
 {
@@ -1661,7 +1723,7 @@ static void cm_update_powered(struct modem_data *modem,
 	cm_set_powered(modem, TRUE);
 }
 
-static gboolean cm_changed(DBusConnection *connection, DBusMessage *message,
+static gboolean cm_changed(DBusConnection *conn, DBusMessage *message,
 				void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -1716,7 +1778,7 @@ static void cdma_cm_update_settings(struct modem_data *modem,
 	extract_ipv4_settings(value, modem->context);
 }
 
-static gboolean cdma_cm_changed(DBusConnection *connection,
+static gboolean cdma_cm_changed(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -1822,7 +1884,7 @@ static void sim_update_imsi(struct modem_data *modem,
 	modem->imsi = g_strdup(imsi);
 }
 
-static gboolean sim_changed(DBusConnection *connection, DBusMessage *message,
+static gboolean sim_changed(DBusConnection *conn, DBusMessage *message,
 				void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -2000,7 +2062,7 @@ static void modem_update_interfaces(struct modem_data *modem,
 	}
 }
 
-static gboolean modem_changed(DBusConnection *connection, DBusMessage *message,
+static gboolean modem_changed(DBusConnection *conn, DBusMessage *message,
 				void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
@@ -2193,7 +2255,7 @@ static void remove_modem(gpointer data)
 	g_free(modem);
 }
 
-static gboolean modem_added(DBusConnection *connection,
+static gboolean modem_added(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
 	DBusMessageIter iter, properties;
@@ -2214,7 +2276,7 @@ static gboolean modem_added(DBusConnection *connection,
 	return TRUE;
 }
 
-static gboolean modem_removed(DBusConnection *connection,
+static gboolean modem_removed(DBusConnection *conn,
 				DBusMessage *message, void *user_data)
 {
 	DBusMessageIter iter;
@@ -2391,7 +2453,7 @@ static int network_disconnect(struct connman_network *network)
 }
 
 static struct connman_network_driver network_driver = {
-	.name		= "network",
+	.name		= "cellular",
 	.type		= CONNMAN_NETWORK_TYPE_CELLULAR,
 	.probe		= network_probe,
 	.remove		= network_remove,
@@ -2491,68 +2553,71 @@ static int ofono_init(void)
 					OFONO_SERVICE, ofono_connect,
 					ofono_disconnect, NULL, NULL);
 
-	modem_added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
-						OFONO_MANAGER_INTERFACE,
+	modem_added_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE,
+						NULL, OFONO_MANAGER_INTERFACE,
 						MODEM_ADDED,
 						modem_added,
 						NULL, NULL);
 
-	modem_removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	modem_removed_watch = g_dbus_add_signal_watch(connection,
+						OFONO_SERVICE, NULL,
 						OFONO_MANAGER_INTERFACE,
 						MODEM_REMOVED,
 						modem_removed,
 						NULL, NULL);
 
-	modem_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	modem_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE, NULL,
 						OFONO_MODEM_INTERFACE,
 						PROPERTY_CHANGED,
 						modem_changed,
 						NULL, NULL);
 
-	cm_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	cm_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE, NULL,
 						OFONO_CM_INTERFACE,
 						PROPERTY_CHANGED,
 						cm_changed,
 						NULL, NULL);
 
-	sim_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	sim_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE, NULL,
 						OFONO_SIM_INTERFACE,
 						PROPERTY_CHANGED,
 						sim_changed,
 						NULL, NULL);
 
-	context_added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	context_added_watch = g_dbus_add_signal_watch(connection,
+						OFONO_SERVICE, NULL,
 						OFONO_CM_INTERFACE,
 						CONTEXT_ADDED,
 						cm_context_added,
 						NULL, NULL);
 
-	context_removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	context_removed_watch = g_dbus_add_signal_watch(connection,
+						OFONO_SERVICE, NULL,
 						OFONO_CM_INTERFACE,
 						CONTEXT_REMOVED,
 						cm_context_removed,
 						NULL, NULL);
 
-	context_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
-						OFONO_CONTEXT_INTERFACE,
+	context_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE,
+						NULL, OFONO_CONTEXT_INTERFACE,
 						PROPERTY_CHANGED,
 						context_changed,
 						NULL, NULL);
 
-	netreg_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	netreg_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE, NULL,
 						OFONO_NETREG_INTERFACE,
 						PROPERTY_CHANGED,
 						netreg_changed,
 						NULL, NULL);
 
-	cdma_cm_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
-						OFONO_CDMA_CM_INTERFACE,
+	cdma_cm_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE,
+						NULL, OFONO_CDMA_CM_INTERFACE,
 						PROPERTY_CHANGED,
 						cdma_cm_changed,
 						NULL, NULL);
 
-	cdma_netreg_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
-						OFONO_CDMA_NETREG_INTERFACE,
+	cdma_netreg_watch = g_dbus_add_signal_watch(connection, OFONO_SERVICE,
+						NULL, OFONO_CDMA_NETREG_INTERFACE,
 						PROPERTY_CHANGED,
 						cdma_netreg_changed,
 						NULL, NULL);
