@@ -47,7 +47,7 @@ static void agent_free(void)
 	agent_path = NULL;
 }
 
-static void agent_disconnect(DBusConnection *connection, void *data)
+static void agent_disconnect(DBusConnection *conn, void *data)
 {
 	DBG("data %p", data);
 
@@ -85,6 +85,26 @@ int __connman_agent_unregister(const char *sender, const char *path)
 	return 0;
 }
 
+static connman_bool_t check_reply_has_dict(DBusMessage *reply)
+{
+	const char *signature = DBUS_TYPE_ARRAY_AS_STRING
+		DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+		DBUS_TYPE_STRING_AS_STRING
+		DBUS_TYPE_VARIANT_AS_STRING
+		DBUS_DICT_ENTRY_END_CHAR_AS_STRING;
+
+	if (dbus_message_has_signature(reply, signature) == TRUE)
+		return TRUE;
+
+	connman_warn("Reply %s to %s from %s has wrong signature %s",
+			signature,
+			dbus_message_get_interface(reply),
+			dbus_message_get_sender(reply),
+			dbus_message_get_signature(reply));
+
+	return FALSE;
+}
+
 struct request_input_reply {
 	struct connman_service *service;
 	authentication_cb_t callback;
@@ -111,6 +131,9 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 		goto done;
 	}
 
+	if (check_reply_has_dict(reply) == FALSE)
+		goto done;
+
 	values_received = TRUE;
 
 	dbus_message_iter_init(reply, &iter);
@@ -311,10 +334,70 @@ static void request_input_append_password(DBusMessageIter *iter,
 				DBUS_TYPE_STRING, &str);
 }
 
+struct previous_passphrase_data {
+	const char *passphrase;
+	const char *type;
+};
+
+static void request_input_append_previouspassphrase(DBusMessageIter *iter,
+							void *user_data)
+{
+	struct previous_passphrase_data *data = user_data;
+	const char *requirement = "informational";
+
+	connman_dbus_dict_append_basic(iter, "Type",
+				DBUS_TYPE_STRING, &data->type);
+
+	connman_dbus_dict_append_basic(iter, "Requirement",
+				DBUS_TYPE_STRING, &requirement);
+
+	connman_dbus_dict_append_basic(iter, "Value",
+				DBUS_TYPE_STRING, &data->passphrase);
+}
+
+static void previous_passphrase_handler(DBusMessageIter *iter,
+					struct connman_service *service)
+{
+	enum connman_service_security security;
+	struct previous_passphrase_data data;
+	struct connman_network *network;
+
+	network = __connman_service_get_network(service);
+	data.passphrase = connman_network_get_string(network, "WiFi.PinWPS");
+
+	if (connman_network_get_bool(network, "WiFi.UseWPS") == TRUE &&
+						data.passphrase != NULL) {
+		data.type = "wpspin";
+	} else {
+		data.passphrase = __connman_service_get_passphrase(service);
+		if (data.passphrase == NULL)
+			return;
+
+		security = __connman_service_get_security(service);
+		switch (security) {
+		case CONNMAN_SERVICE_SECURITY_WEP:
+			data.type = "wep";
+			break;
+		case CONNMAN_SERVICE_SECURITY_PSK:
+			data.type  = "psk";
+			break;
+		/*
+		 * This should never happen: no passphrase is set if security
+		 * is not one of the above. */
+		default:
+			break;
+		}
+	}
+
+	connman_dbus_dict_append_dict(iter, "PreviousPassphrase",
+			request_input_append_previouspassphrase, &data);
+}
+
 static void request_input_login_reply(DBusPendingCall *call, void *user_data)
 {
 	struct request_input_reply *username_password_reply = user_data;
 	const char *error = NULL;
+	connman_bool_t values_received = FALSE;
 	char *username = NULL;
 	char *password = NULL;
 	char *key;
@@ -326,6 +409,11 @@ static void request_input_login_reply(DBusPendingCall *call, void *user_data)
 		goto done;
 	}
 
+	if (check_reply_has_dict(reply) == FALSE)
+		goto done;
+
+	values_received = TRUE;
+
 	dbus_message_iter_init(reply, &iter);
 	dbus_message_iter_recurse(&iter, &dict);
 	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
@@ -359,7 +447,7 @@ static void request_input_login_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	username_password_reply->callback(username_password_reply->service,
-					TRUE, NULL, 0,
+					values_received, NULL, 0,
 					username, password,
 					FALSE, NULL, error,
 					username_password_reply->user_data);
@@ -412,6 +500,8 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 			CONNMAN_SERVICE_SECURITY_NONE) {
 		connman_dbus_dict_append_dict(&dict, "Passphrase",
 					request_input_append_passphrase, service);
+
+		previous_passphrase_handler(&dict, service);
 	}
 
 	if (__connman_service_wps_enabled(service) == TRUE) {
@@ -427,8 +517,9 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message,
-						&call, -1) == FALSE) {
+	if (dbus_connection_send_with_reply(connection, message, &call,
+					connman_timeout_input_request())
+			== FALSE) {
 		dbus_message_unref(message);
 		g_free(passphrase_reply);
 		return -ESRCH;
@@ -449,7 +540,7 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 
 	dbus_message_unref(message);
 
-	return -EIO;
+	return -EINPROGRESS;
 }
 
 int __connman_agent_request_login_input(struct connman_service *service,
@@ -493,8 +584,9 @@ int __connman_agent_request_login_input(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message,
-							&call, -1) == FALSE) {
+	if (dbus_connection_send_with_reply(connection, message, &call,
+					connman_timeout_input_request())
+			== FALSE) {
 		dbus_message_unref(message);
 		g_free(username_password_reply);
 		return -ESRCH;
@@ -515,7 +607,7 @@ int __connman_agent_request_login_input(struct connman_service *service,
 
 	dbus_message_unref(message);
 
-	return -EIO;
+	return -EINPROGRESS;
 }
 
 struct request_browser_reply_data {
@@ -581,8 +673,9 @@ int __connman_agent_request_browser(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message,
-							&call, -1) == FALSE) {
+	if (dbus_connection_send_with_reply(connection, message, &call,
+					connman_timeout_browser_launch())
+			== FALSE) {
 		dbus_message_unref(message);
 		g_free(browser_reply_data);
 		return -ESRCH;
@@ -603,7 +696,7 @@ int __connman_agent_request_browser(struct connman_service *service,
 
 	dbus_message_unref(message);
 
-	return -EIO;
+	return -EINPROGRESS;
 }
 
 struct report_error_data {
@@ -669,8 +762,9 @@ int __connman_agent_report_error(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message,
-						&call, -1) == FALSE) {
+	if (dbus_connection_send_with_reply(connection, message, &call,
+					connman_timeout_input_request())
+			== FALSE) {
 		dbus_message_unref(message);
 		g_free(report_error);
 		return -ESRCH;
@@ -689,7 +783,7 @@ int __connman_agent_report_error(struct connman_service *service,
 				report_error, NULL);
 	dbus_message_unref(message);
 
-	return -EIO;
+	return -EINPROGRESS;
 }
 
 int __connman_agent_init(void)
