@@ -54,7 +54,6 @@ struct connman_device {
 	char *ident;
 	char *path;
 	char *devname;
-	int phyindex;
 	int index;
 	guint pending_timeout;
 
@@ -238,7 +237,7 @@ int __connman_device_disable(struct connman_device *device)
 
 	if (device->network) {
 		struct connman_service *service =
-			__connman_service_lookup_from_network(device->network);
+			connman_service_lookup_from_network(device->network);
 
 		if (service != NULL)
 			__connman_service_disconnect(service);
@@ -247,12 +246,7 @@ int __connman_device_disable(struct connman_device *device)
 	}
 
 	err = device->driver->disable(device);
-	if (err == 0) {
-		connman_device_set_powered(device, FALSE);
-		goto done;
-	}
-
-	if (err == -EALREADY) {
+	if (err == 0 || err == -EALREADY) {
 		connman_device_set_powered(device, FALSE);
 		goto done;
 	}
@@ -428,8 +422,6 @@ struct connman_device *connman_device_create(const char *node,
 	device->type = type;
 	device->name = g_strdup(type2description(device->type));
 
-	device->phyindex = -1;
-
 	device->networks = g_hash_table_new_full(g_str_hash, g_str_equal,
 						g_free, free_network);
 
@@ -519,17 +511,6 @@ int connman_device_get_index(struct connman_device *device)
 	return device->index;
 }
 
-int __connman_device_get_phyindex(struct connman_device *device)
-{
-	return device->phyindex;
-}
-
-void __connman_device_set_phyindex(struct connman_device *device,
-							int phyindex)
-{
-	device->phyindex = phyindex;
-}
-
 /**
  * connman_device_set_interface:
  * @device: device structure
@@ -598,13 +579,12 @@ int connman_device_set_powered(struct connman_device *device,
 
 	type = __connman_device_get_service_type(device);
 
-	if (device->powered == TRUE)
-		__connman_technology_enabled(type);
-	else
+	if (device->powered == FALSE) {
 		__connman_technology_disabled(type);
-
-	if (powered == FALSE)
 		return 0;
+	}
+
+	__connman_technology_enabled(type);
 
 	connman_device_set_disconnected(device, FALSE);
 	device->scanning = FALSE;
@@ -710,8 +690,6 @@ connman_bool_t connman_device_get_scanning(struct connman_device *device)
 
 void connman_device_reset_scanning(struct connman_device *device)
 {
-	device->scanning = FALSE;
-
 	g_hash_table_foreach(device->networks,
 				mark_network_available, NULL);
 }
@@ -1053,6 +1031,38 @@ struct connman_device *__connman_device_find_device(
 	return NULL;
 }
 
+/**
+ * connman_device_set_regdom
+ * @device: device structure
+ * @alpha2: string representing regulatory domain
+ *
+ * Set regulatory domain on device basis
+ */
+int connman_device_set_regdom(struct connman_device *device,
+						const char *alpha2)
+{
+	if (device->driver == NULL || device->driver->set_regdom == NULL)
+		return -ENOTSUP;
+
+	if (device->powered == FALSE)
+		return -EINVAL;
+
+	return device->driver->set_regdom(device, alpha2);
+}
+
+/**
+ * connman_device_regdom_notify
+ * @device: device structure
+ * @alpha2: string representing regulatory domain
+ *
+ * Notify on setting regulatory domain on device basis
+ */
+void connman_device_regdom_notify(struct connman_device *device,
+					int result, const char *alpha2)
+{
+	__connman_technology_notify_regdom_by_device(device, result, alpha2);
+}
+
 int __connman_device_request_scan(enum connman_service_type type)
 {
 	connman_bool_t success = FALSE;
@@ -1102,7 +1112,8 @@ int __connman_device_request_scan(enum connman_service_type type)
 
 int __connman_device_request_hidden_scan(struct connman_device *device,
 				const char *ssid, unsigned int ssid_len,
-				const char *identity, const char *passphrase)
+				const char *identity, const char *passphrase,
+				void *user_data)
 {
 	DBG("device %p", device);
 
@@ -1114,12 +1125,13 @@ int __connman_device_request_hidden_scan(struct connman_device *device,
 		return -EALREADY;
 
 	return device->driver->scan_hidden(device, ssid, ssid_len,
-					identity, passphrase);
+					identity, passphrase, user_data);
 }
 
 connman_bool_t __connman_device_isfiltered(const char *devname)
 {
 	char **pattern;
+	char **blacklisted_interfaces;
 
 	if (device_filter == NULL)
 		goto nodevice;
@@ -1138,7 +1150,7 @@ nodevice:
 	}
 
 	if (nodevice_filter == NULL)
-		return FALSE;
+		goto list;
 
 	for (pattern = nodevice_filter; *pattern; pattern++) {
 		if (g_pattern_match_simple(*pattern, devname) == TRUE) {
@@ -1147,9 +1159,70 @@ nodevice:
 		}
 	}
 
+list:
+	blacklisted_interfaces =
+		connman_setting_get_string_list("NetworkInterfaceBlacklist");
+	if (blacklisted_interfaces == NULL)
+		return FALSE;
+
+	for (pattern = blacklisted_interfaces; *pattern; pattern++) {
+		if (g_str_has_prefix(devname, *pattern) == TRUE) {
+			DBG("ignoring device %s (blacklist)", devname);
+			return TRUE;
+		}
+	}
+
 	return FALSE;
 }
 
+static void cleanup_devices(void)
+{
+	/*
+	 * Check what interfaces are currently up and if connman is
+	 * suppose to handle the interface, then cleanup the mess
+	 * related to that interface. There might be weird routes etc
+	 * that are related to that interface and that might confuse
+	 * connmand. So in this case we just turn the interface down
+	 * so that kernel removes routes/addresses automatically and
+	 * then proceed the startup.
+	 *
+	 * Note that this cleanup must be done before rtnl/detect code
+	 * has activated interface watches.
+	 */
+
+	char **interfaces;
+	int i;
+
+	interfaces = __connman_inet_get_running_interfaces();
+
+	if (interfaces == NULL)
+		return;
+
+	for (i = 0; interfaces[i] != NULL; i++) {
+		connman_bool_t filtered;
+		int index;
+
+		filtered = __connman_device_isfiltered(interfaces[i]);
+		if (filtered == TRUE)
+			continue;
+
+		index = connman_inet_ifindex(interfaces[i]);
+		if (index < 0)
+			continue;
+
+		DBG("cleaning up %s index %d", interfaces[i], index);
+
+		connman_inet_ifdown(index);
+
+		/*
+		 * ConnMan will turn the interface UP automatically so
+		 * no need to do it here.
+		 */
+	}
+
+	g_strfreev(interfaces);
+}
+
 int __connman_device_init(const char *device, const char *nodevice)
 {
 	DBG("");
@@ -1160,6 +1233,8 @@ int __connman_device_init(const char *device, const char *nodevice)
 	if (nodevice != NULL)
 		nodevice_filter = g_strsplit(nodevice, ",", -1);
 
+	cleanup_devices();
+
 	return 0;
 }
 
