@@ -42,6 +42,8 @@
 #define DEFAULT_INPUT_REQUEST_TIMEOUT 120 * 1000
 #define DEFAULT_BROWSER_LAUNCH_TIMEOUT 300 * 1000
 
+#define CONFIGMAINFILE CONFIGDIR "/main.conf"
+
 static char *default_auto_connect[] = {
 	"wifi",
 	"ethernet",
@@ -65,6 +67,7 @@ static struct {
 	unsigned int timeout_inputreq;
 	unsigned int timeout_browserlaunch;
 	char **blacklisted_interfaces;
+	connman_bool_t allow_hostname_updates;
 } connman_settings  = {
 	.bg_scan = TRUE,
 	.pref_timeservers = NULL,
@@ -74,6 +77,7 @@ static struct {
 	.timeout_inputreq = DEFAULT_INPUT_REQUEST_TIMEOUT,
 	.timeout_browserlaunch = DEFAULT_BROWSER_LAUNCH_TIMEOUT,
 	.blacklisted_interfaces = NULL,
+	.allow_hostname_updates = TRUE,
 };
 
 static GKeyFile *load_config(const char *file)
@@ -241,6 +245,26 @@ static void parse_config(GKeyFile *config)
 			g_strdupv(default_blacklist);
 
 	g_clear_error(&error);
+
+	boolean = g_key_file_get_boolean(config, "General",
+					"AllowHostnameUpdates",
+					&error);
+	if (error == NULL)
+		connman_settings.allow_hostname_updates = boolean;
+
+	g_clear_error(&error);
+}
+
+static int config_init(const char *file)
+{
+	GKeyFile *config;
+
+	config = load_config(file);
+	parse_config(config);
+	if (config != NULL)
+		g_key_file_free(config);
+
+	return 0;
 }
 
 static GMainLoop *main_loop = NULL;
@@ -322,6 +346,7 @@ static void disconnect_callback(DBusConnection *conn, void *user_data)
 	g_main_loop_quit(main_loop);
 }
 
+static gchar *option_config = NULL;
 static gchar *option_debug = NULL;
 static gchar *option_device = NULL;
 static gchar *option_plugin = NULL;
@@ -330,7 +355,6 @@ static gchar *option_noplugin = NULL;
 static gchar *option_wifi = NULL;
 static gboolean option_detach = TRUE;
 static gboolean option_dnsproxy = TRUE;
-static gboolean option_compat = FALSE;
 static gboolean option_version = FALSE;
 
 static gboolean parse_debug(const char *key, const char *value,
@@ -345,6 +369,9 @@ static gboolean parse_debug(const char *key, const char *value,
 }
 
 static GOptionEntry options[] = {
+	{ "config", 'c', 0, G_OPTION_ARG_STRING, &option_config,
+				"Load the specified configuration file "
+				"instead of " CONFIGMAINFILE, "FILE" },
 	{ "debug", 'd', G_OPTION_FLAG_OPTIONAL_ARG,
 				G_OPTION_ARG_CALLBACK, parse_debug,
 				"Specify debug options to enable", "DEBUG" },
@@ -364,8 +391,6 @@ static GOptionEntry options[] = {
 	{ "nodnsproxy", 'r', G_OPTION_FLAG_REVERSE,
 				G_OPTION_ARG_NONE, &option_dnsproxy,
 				"Don't enable DNS Proxy" },
-	{ "compat", 'c', 0, G_OPTION_ARG_NONE, &option_compat,
-				"(obsolete)" },
 	{ "version", 'v', 0, G_OPTION_ARG_NONE, &option_version,
 				"Show version information and exit" },
 	{ NULL },
@@ -388,6 +413,9 @@ connman_bool_t connman_setting_get_bool(const char *key)
 	if (g_str_equal(key, "BackgroundScanning") == TRUE)
 		return connman_settings.bg_scan;
 
+	if (g_str_equal(key, "AllowHostnameUpdates") == TRUE)
+		return connman_settings.allow_hostname_updates;
+
 	return FALSE;
 }
 
@@ -430,7 +458,6 @@ int main(int argc, char *argv[])
 	GError *error = NULL;
 	DBusConnection *conn;
 	DBusError err;
-	GKeyFile *config;
 	guint signal;
 
 #ifdef NEED_THREADS
@@ -507,10 +534,10 @@ int main(int argc, char *argv[])
 
 	__connman_dbus_init(conn);
 
-	config = load_config(CONFIGDIR "/main.conf");
-	parse_config(config);
-	if (config != NULL)
-		g_key_file_free(config);
+	if (option_config == NULL)
+		config_init(CONFIGMAINFILE);
+	else
+		config_init(option_config);
 
 	__connman_storage_migrate();
 	__connman_technology_init();
@@ -550,6 +577,7 @@ int main(int argc, char *argv[])
 	__connman_wispr_init();
 	__connman_rfkill_init();
 
+	g_free(option_config);
 	g_free(option_device);
 	g_free(option_plugin);
 	g_free(option_nodevice);
